from pathlib import Path

from muextensions.connector.docutils import plantuml, ditaa


BUILD_PATH = '_generated'


def register(args):
    generate_dir = Path(args.targetdir).joinpath(BUILD_PATH)
    plantuml.register(
        target_dir=generate_dir,
        base_uri=BUILD_PATH,
        create_dir=True)
    ditaa.register(
        target_dir=generate_dir,
        base_uri=BUILD_PATH,
        create_dir=True)



# pylint: disable=missing-docstring

import hashlib
import itertools
import logging
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from . import execute, ExecutionError


LOG = logging.getLogger(__name__)


class PlantUmlWrapper():
    def __init__(self,
                 uml_body,
                 output_format='svg',
                 encoding='utf-8',
                 timeout=30):
        try:
            iter(uml_body)
        except TypeError:
            raise TypeError('"uml_body" must be iterable')

        if output_format not in ('png', 'svg'):
            raise ValueError('"output_format" must be one of: png, svg')

        self._uml_body = uml_body
        self._format = output_format
        self._encoding = encoding
        self._timeout = timeout
        self._log = logging.getLogger(type(self).__name__)

    def write(self, output_file):
        with TemporaryDirectory(prefix='muextensions-plantuml-') as work_dir:
            self._process(work_dir, output_file)

    def hashcode(self):
        hasher = hashlib.md5()
        for item in self._uml_iterator():
            hasher.update(item)
        return hasher.hexdigest()

    def _process(self, work_dir, output_file):
        base_name = self.hashcode()
        puml_path = Path(work_dir, base_name).with_suffix('.puml')
        self._write_uml_file(puml_path)
        output_path = self._execute_plantuml(puml_path)
        self._log.debug('Moving "%s" to "%s".', output_path, output_file)
        shutil.move(str(output_path), str(output_file))

    def _execute_plantuml(self, puml_path):
        option = '-t{}'.format(self._format)
        execute(['plantuml', option, puml_path])
        output_path = puml_path.with_suffix('.' + self._format)
        if not output_path.exists():
            raise ExecutionError('Failed to create "{}"'.format(output_path))
        return output_path

    def _uml_iterator(self):
        data = itertools.chain(['@startuml'], self._uml_body, ['@enduml'])
        return ('{}\n'.format(line).encode(self._encoding) for line in data)

    def _write_uml_file(self, puml_path):
        with puml_path.open('bw+') as handle:
            for line in self._uml_iterator():
                handle.write(line)



from pathlib import Path
import os

from pelican import signals, logger

from muextensions.connector.docutils import plantuml, ditaa


BUILD_PATH = '_generated'


def pelican_init(pelicanobj):
    target_dir = os.path.join('images', BUILD_PATH)
    base_uri = '/' + target_dir
    output_path = Path(pelicanobj.settings['OUTPUT_PATH'])
    generate_dir = output_path.joinpath(target_dir)
    logger.info('Registered "%s" with working directory "%s" '
                'and a base URL of "%s"', __name__, generate_dir, base_uri)
    plantuml.register(
        target_dir=generate_dir,
        base_uri=base_uri,
        create_dir=True)
    ditaa.register(
        target_dir=generate_dir,
        base_uri=base_uri,
        create_dir=True)


def register():
    signals.initialized.connect(pelican_init)



from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, patch

from muextensions.contrib import hovercraft


@patch('muextensions.contrib.hovercraft.ditaa')
@patch('muextensions.contrib.hovercraft.plantuml')
def test_register(plantuml_mock, ditaa_mock):
    args = MagicMock()
    type(args).targetdir = PropertyMock(return_value='some/path/here')

    hovercraft.register(args)

    plantuml_mock.register.assert_called_with(
        target_dir=Path('some/path/here/_generated'),
        base_uri='_generated',
        create_dir=True)
    ditaa_mock.register.assert_called_with(
        target_dir=Path('some/path/here/_generated'),
        base_uri='_generated',
        create_dir=True)



from docutils.parsers.rst.directives.images import Image
from docutils.parsers.rst.directives import register_directive
from docutils.parsers.rst import directives

from muextensions.executor.ditaa import DitaaWrapper
from . import Connector


class DitaaImageDirectiveBase(Connector, Image):
    required_arguments = 0
    has_content = True
    option_spec = Image.option_spec.copy()
    option_spec['basename'] = directives.unchanged
    option_spec['format'] = lambda _: directives.choice(_, ('png', 'svg'))

    # FIXME - Move to Connector.run()
    def run(self):
        image_format = self.options.get('format', self.default_image_format)

        # FIXME - Move to DitaaImageDirectiveBase.get_executor()
        executor = DitaaWrapper(self.content, output_format=image_format)

        basename = self.options.get('basename', executor.hashcode())
        target_path = self.target_dir.joinpath(basename) \
            .with_suffix('.' + image_format)
        if not target_path.parent.exists() and self.create_dirs:
            target_path.parent.mkdir(parents=True)

        # FIXME - Add docutils error handling!!!
        executor.write(target_path)

        if self.base_uri is not None:
            uri = '/'.join([self.base_uri, target_path.name])
        else:
            uri = str(target_path)

        self.arguments.append(uri)
        return super().run()


def register(target_dir, base_uri=None, default_image_format='png',
             create_dir=False):
    register_directive(
        'ditaa-image',
        build_ditaa_image_directive(
            target_dir, base_uri, default_image_format, create_dir))


def build_ditaa_image_directive(
        target_dir, base_uri, default_image_format, create_dir):
    return type('DitaaImageDirective',
                (DitaaImageDirectiveBase,),
                {'_target_dir': target_dir,
                 '_base_uri': base_uri,
                 '_default_image_format': default_image_format,
                 '_create_dirs': create_dir})



# pylint: disable=missing-docstring

import logging
import subprocess


LOG = logging.getLogger(__name__)


class ExecutionError(Exception):
    pass


class Timeout(ExecutionError):
    pass


def execute(command, timeout=30):
    try:
        LOG.debug('Executinge "%s".', command)
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        _kill_process(process)
        raise Timeout('Call to PlantUML took too long')
    except subprocess.SubprocessError:
        _kill_process(process)
        raise ExecutionError('Call to PlantUML failed with an exception')
    if process.returncode != 0:
        _log_process_failed(process, stdout, stderr)
        raise ExecutionError('Call to PlantUML failed')
    _log_process_sucess(process, stdout, stderr)


def _log_process_sucess(process, stdout, stderr):
    _log_process(True, process, stdout, stderr)


def _log_process_failed(process, stdout, stderr):
    _log_process(False, process, stdout, stderr)


def _kill_process(process):
    process.kill()
    stdout, stderr = process.communicate()
    _log_process_failed(process, stdout, stderr)


def _log_process(success, process, stdout, stderr):
    comment = 'succeeded' if success else 'failed'
    if success:
        comment = 'succeeded'
        level = logging.DEBUG
    else:
        comment = 'failed'
        level = logging.ERROR
    LOG.log(level, 'Command "%s" %s with a return code of "%d". '
            'STDOUT: %s STDERR: %s',
            process.args, comment, process.returncode, stdout, stderr)



class Connector():
    @property
    def target_dir(self):
        # pylint: disable=no-member
        return self._target_dir

    @property
    def base_uri(self):
        # pylint: disable=no-member
        return self._base_uri

    @property
    def default_image_format(self):
        # pylint: disable=no-member
        return self._default_image_format

    @property
    def create_dirs(self):
        # pylint: disable=no-member
        return self._create_dirs



from pathlib import Path
from unittest.mock import patch, sentinel, MagicMock
import pytest
from bs4 import BeautifulSoup

from docutils.core import publish_file
from docutils.parsers.rst.directives.images import Image

from muextensions.connector.docutils import ditaa

from ...testhelper import preserve_docutils_directives


RESOURCES_DIR = Path(__file__).with_suffix('')
REST_SAMPLE_FILE = RESOURCES_DIR / 'sample-default.rst'


# -----------------------------------------------------------------------------
# register()
# -----------------------------------------------------------------------------
@patch('muextensions.connector.docutils.ditaa.register_directive')
@patch('muextensions.connector.docutils.ditaa.build_ditaa_image_directive')
def test_register_defaults(build_mock, register_directive_mock, tmp_path):
    # pylint: disable=protected-access
    type(build_mock).return_value = sentinel.clazz
    ditaa.register(tmp_path)
    build_mock.assert_called_once_with(tmp_path, None, 'png', False)
    register_directive_mock.assert_called_once_with(
        'ditaa-image', sentinel.clazz)


@patch('muextensions.connector.docutils.ditaa.register_directive')
@patch('muextensions.connector.docutils.ditaa.build_ditaa_image_directive')
def test_register_modified(build_mock, register_directive_mock, tmp_path):
    type(build_mock).return_value = sentinel.clazz
    ditaa.register(tmp_path, 'some/base/uri', 'svg', True)
    build_mock.assert_called_once_with(tmp_path, 'some/base/uri', 'svg', True)
    register_directive_mock.assert_called_once_with(
        'ditaa-image', sentinel.clazz)


# -----------------------------------------------------------------------------
# build_ditaa_image_directive()
# -----------------------------------------------------------------------------
def test_build_ditaa_image_directive():
    # pylint: disable=protected-access
    clazz = ditaa.build_ditaa_image_directive(
        sentinel.target_dir, sentinel.base_uri, sentinel.default_image_format,
        sentinel.create_dir)
    assert issubclass(clazz, Image)
    assert clazz.__module__ == 'muextensions.connector.docutils.ditaa'
    assert clazz.__name__ == 'DitaaImageDirective'
    assert clazz._target_dir is sentinel.target_dir
    assert clazz._base_uri is sentinel.base_uri
    assert clazz._default_image_format is sentinel.default_image_format
    assert clazz._create_dirs is sentinel.create_dir


# -----------------------------------------------------------------------------
# DitaaImageDirectiveBase.run()
# -----------------------------------------------------------------------------
@preserve_docutils_directives
@pytest.mark.parametrize('image_format', ('png', 'svg'))
def test_image_directive_run_baseline(
        ditaa_wrapper_class_mock, image_format, integration_output_path):
    # pylint: disable=redefined-outer-name
    # arrange
    ditaa.register(integration_output_path, default_image_format=image_format)
    output = integration_output_path.joinpath('sample.html')
    # apply
    publish_file(source_path=str(REST_SAMPLE_FILE),
                 destination_path=str(output),
                 writer_name='html5')
    # assert
    expected_path = integration_output_path \
        .joinpath('fake_hash_code').with_suffix('.' + image_format)
    ditaa_wrapper_class_mock.instantiated_obj_mock \
        .write.assert_called_once_with(expected_path)
    with output.open(encoding='utf8') as data:
        soup = BeautifulSoup(data, 'html.parser')
    images = soup.find_all('img')
    assert len(images) == 1
    assert images[0]['alt'] == images[0]['src']
    assert Path(images[0]['src']).suffix == '.' + image_format
    assert str(Path(images[0]['src']).parent) == str(integration_output_path)


@preserve_docutils_directives
def test_image_directive_run_default_base_uri(
        ditaa_wrapper_class_mock, integration_output_path):
    # pylint: disable=redefined-outer-name,unused-argument
    ditaa.register(integration_output_path, '.')
    output = integration_output_path.joinpath('sample.html')
    publish_file(source_path=str(REST_SAMPLE_FILE),
                 destination_path=str(output),
                 writer_name='html5')
    with output.open(encoding='utf8') as data:
        soup = BeautifulSoup(data, 'html.parser')
    images = soup.find_all('img')
    assert images[0]['src'] == './fake_hash_code.png'


@preserve_docutils_directives
def test_image_directive_run_custom_base_uri(
        ditaa_wrapper_class_mock, integration_output_path):
    # pylint: disable=redefined-outer-name,unused-argument
    base_uri = './some/dir'
    images_path = integration_output_path.joinpath(base_uri)
    ditaa.register(images_path, base_uri, create_dir=True)
    html_output = integration_output_path.joinpath('sample.html')
    publish_file(source_path=str(REST_SAMPLE_FILE),
                 destination_path=str(html_output),
                 writer_name='html5')
    with html_output.open(encoding='utf8') as data:
        soup = BeautifulSoup(data, 'html.parser')
    images = soup.find_all('img')
    assert images[0]['src'] == './some/dir/fake_hash_code.png'


# -----------------------------------------------------------------------------
# Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def ditaa_wrapper_class_mock():
    # arrange
    with patch('muextensions.connector.docutils.ditaa.DitaaWrapper') as clazz:
        obj_mock = MagicMock()
        obj_mock.hashcode.return_value = 'fake_hash_code'
        type(clazz).return_value = obj_mock
        clazz.instantiated_obj_mock = obj_mock
        yield clazz



from unittest.mock import patch, ANY, call

from muextensions.connector.docutils import plantuml


@patch('muextensions.connector.docutils.plantuml.directives')
def test_register_defaults(directives_mock):
    # pylint: disable=protected-access
    plantuml.register('a/target/directory')
    directives_mock.register_directive.assert_has_calls((
        call('plantuml-image', ANY),
        call('uml', ANY)))
    directive_class = directives_mock.register_directive.call_args[0][1]
    assert directive_class._target_dir == 'a/target/directory'
    assert directive_class._base_uri is None
    assert not directive_class._create_dirs


@patch('muextensions.connector.docutils.plantuml.directives')
def test_register_custom_args(directives_mock):
    # pylint: disable=protected-access
    plantuml.register('another/target/directory', 'some/base', True)
    directive_class = directives_mock.register_directive.call_args[0][1]
    assert directive_class._target_dir == 'another/target/directory'
    assert directive_class._base_uri == 'some/base'
    assert directive_class._create_dirs


@patch('muextensions.connector.docutils.plantuml.directives')
def test_directive_defaults(directives_mock):
    # arrange
    plantuml.register('foo/bar')
    directive_class = directives_mock.register_directive.call_args[0][1]
    test_class = type('Foo', (directive_class,), {'__init__': lambda _: None})
    # act
    directive = test_class()
    # assert
    assert directive.target_dir == 'foo/bar'
    assert directive.base_uri is None
    assert not directive.create_dirs


@patch('muextensions.connector.docutils.plantuml.directives')
def test_directive_custom_args(directives_mock):
    # arrange
    plantuml.register('here/is/a/directory', 'a/base/2', True)
    directive_class = directives_mock.register_directive.call_args[0][1]
    test_class = type('Foo', (directive_class,), {'__init__': lambda _: None})
    # act
    directive = test_class()
    # assert
    assert directive.target_dir == 'here/is/a/directory'
    assert directive.base_uri == 'a/base/2'
    assert directive.create_dirs



from pathlib import Path
from unittest.mock import patch, sentinel

from muextensions.contrib import pelican


@patch('muextensions.contrib.pelican.ditaa')
@patch('muextensions.contrib.pelican.plantuml')
def test_pelican_init(plantuml_mock, ditaa_mock):
    pelican_obj = sentinel.pelican_obj
    pelican_obj.settings = {'OUTPUT_PATH': 'some/target/path'}

    pelican.pelican_init(pelican_obj)

    plantuml_mock.register.assert_called_once_with(
        target_dir=Path('some/target/path/images/_generated'),
        base_uri='/images/_generated',
        create_dir=True)
    ditaa_mock.register.assert_called_once_with(
        target_dir=Path('some/target/path/images/_generated'),
        base_uri='/images/_generated',
        create_dir=True)


@patch('muextensions.contrib.pelican.signals')
def test_register(signals_mock):
    pelican.register()
    signals_mock.initialized.connect.assert_called_with(pelican.pelican_init)



#!/usr/bin/env python
from setuptools import setup, find_packages


def version():
    # pylint: disable=exec-used
    values = {}
    exec(open('muextensions/version.py').read(), values)
    return values['__version__']


def long_description():
    with open('README.rst', 'rb') as handle:
        return handle.read().decode('utf-8')


_TEST_REQUIRE = [
    # coverage 5 is still in alpha.
    'beautifulsoup4',
    'coverage<5',
    'decorator',
    'pytest',
    'pytest-cov',
    # Optional external dependencies needed while testing.
    'docutils',
    'pelican',
]

_CI_REQUIRE = [
    'flake8',
    'pep257',
    'pylint',
    'tox',
]

setup(
    name='muextensions',
    version=version(),
    author='Pedro H.',
    author_email='5179251+pedrohdz@users.noreply.github.com',
    description='Markup extentions',
    long_description=long_description(),
    url='https://github.com/pedrohdz/muextensions',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Documentation',
    ],
    keywords='pelican hovercraft plantuml uml docutils',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    setup_requires=[
        'pytest-runner',
    ],
    install_requires=[
    ],
    tests_require=_TEST_REQUIRE,
    extras_require={
        'ci': _CI_REQUIRE,
        'test': _TEST_REQUIRE,
    },
)



# pylint: disable=missing-docstring
from pathlib import Path

from docutils import nodes
from docutils.parsers.rst import Directive, directives

from muextensions.executor.plantuml import PlantUmlWrapper
from . import Connector


class PlantUmlDocutilsDirectiveBase(Connector, Directive):
    # FIXME - Make like `DitaaImageDirectiveBase`!
    # Potential template for this:
    #   - http://docutils.sourceforge.net/docs/howto/rst-directives.html
    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {
        # image node options
        'align': lambda _: directives.choice(_, ('left', 'center', 'right')),
        'alt': directives.unchanged,
        'class': directives.class_option,
        'height': directives.nonnegative_int,
        'scale': directives.nonnegative_int,
        'width': directives.nonnegative_int,
        # custom options
        'basename': directives.unchanged,
        'format': lambda _: directives.choice(_, ('png', 'svg'))
    }

    def run(self):
        plantuml_format = self.options.get('format', 'png')
        executor = PlantUmlWrapper(
            self.content,
            output_format=plantuml_format
        )

        basename = self.options.get('basename', executor.hashcode())
        file_name = '.'.join([basename, plantuml_format])
        target_path = Path(self.target_dir, file_name)
        if not target_path.parent.exists() and self.create_dirs:
            target_path.parent.mkdir(parents=True)

        executor.write(target_path)

        if self.base_uri is not None:
            self.options['uri'] = '/'.join([self.base_uri, target_path.name])
        else:
            self.options['uri'] = target_path

        image_node = nodes.image(rawsource=self.block_text, **self.options)
        return [image_node]


def register(target_dir, base_uri=None, create_dir=False):
    directive_class = type('PlantUmlDocutilsDirective',
                           (PlantUmlDocutilsDirectiveBase,),
                           {'_target_dir': target_dir,
                            '_base_uri': base_uri,
                            '_create_dirs': create_dir})
    directives.register_directive('plantuml-image', directive_class)

    # TODO - Depricate the `uml` directive in favor of `plantuml-image`.
    directives.register_directive('uml', directive_class)



# pylint: disable=missing-docstring

import io
import hashlib
import logging
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from . import execute


LOG = logging.getLogger(__name__)


class DitaaWrapper():
    def __init__(self,
                 ditaa_body,
                 output_format='svg',
                 encoding='utf-8',
                 timeout=30):
        try:
            iter(ditaa_body)
        except TypeError:
            raise TypeError('"ditaa_body" must be iterable')

        if output_format not in ('png', 'svg'):
            raise ValueError(
                '"output_format" must be one of "png", "svg".  The following '
                'is unknown: {}'.format(output_format))

        self._ditaa_body = ditaa_body
        self._format = output_format
        self._encoding = encoding
        self._timeout = timeout
        self._log = logging.getLogger(type(self).__name__)

    @classmethod
    def from_file(cls, file_path, *args, encoding='utf-8', **kwargs):
        with io.open(str(file_path), 'r', encoding=encoding) as file:
            lines = [l.strip() for l in file.readlines()]
        return cls(lines, encoding=encoding, *args, **kwargs)

    def write(self, output_file):
        with TemporaryDirectory(prefix='muextensions-ditaa-') as work_dir:
            self._process(work_dir, output_file)

    def hashcode(self):
        hasher = hashlib.md5()
        for item in self._ditaa_iterator():
            hasher.update(item)
        return hasher.hexdigest()

    def _process(self, work_dir, output_file):
        base_name = self.hashcode()
        ditaa_path = Path(work_dir, base_name).with_suffix('.ditaa')
        self._write_ditaa_file(ditaa_path)
        output_path = self._execute_ditaa(ditaa_path)
        self._log.debug('Moving "%s" to "%s".', output_path, output_file)
        shutil.move(str(output_path), str(output_file))

    def _execute_ditaa(self, input_path):
        output_path = input_path.with_suffix('.' + self._format)
        command = ['ditaa', str(input_path), str(output_path)]
        command.extend(self._get_options())
        execute(command)
        if not output_path.exists():
            raise FileNotFoundError(
                'Failed to create "{}"'.format(output_path))
        return output_path

    def _get_options(self):
        options = []
        if self._format == 'svg':
            options.append('--svg')
        return options

    def _ditaa_iterator(self):
        return ('{}\n'.format(line).encode(self._encoding) for line in
                self._ditaa_body)

    def _write_ditaa_file(self, ditaa_path):
        with ditaa_path.open('bw+') as handle:
            for line in self._ditaa_iterator():
                handle.write(line)
